/****************************************************************
#
# PART:         idt70v639
# FUNCTION:     128k x 18 Asynchronous Dual-Port Static RAM
# DESIGNER:     Frank Matthews
# DATE:         Wed 5/24/2000 18:12:9
#
*****************************************************************
#
#   Copyright 2000, Integrated Device Technology, Inc.
#   All Rights Reserved
#
***************************************************************
*
* This UNencrypted copy is ONLY delivered under NDA with IDT
* Do NOT use this model if you don't have a NDA in place.
*
* Integrated Device Technology, Inc. ("IDT") hereby grants the 
* user of this Verilog model a non-exclusive, nontransferable 
* license to use this Verilog model under the following terms. 
* The user is granted this license only to use the Verilog 
* model and is not granted rights to sell, copy (except as needed 
* to run the IBIS model), rent, lease or sub-license the Verilog 
* model in whole or in part, or in modified form to anyone. The User 
* may modify the Verilog model to suit its specific applications, 
* but rights to derivative works and such modifications shall belong 
* to IDT. 
* 
* This Verilog model is provided on an "AS IS" basis and 
* IDT makes absolutely no warranty with respect to the information
* contained herein. IDT DISCLAIMS AND CUSTOMER WAIVES ALL
* WARRANTIES, EXPRESS AND IMPLIED, INCLUDING WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
* ENTIRE RISK AS TO QUALITY AND PERFORMANCE IS WITH THE
* USER ACCORDINGLY, IN NO EVENT SHALL IDT BE LIABLE
* FOR ANY DIRECT OR INDIRECT DAMAGES, WHETHER IN CONTRACT OR 
* TORT, INCLUDING ANY LOST PROFITS OR OTHER INCIDENTAL, 
* CONSEQUENTIAL, EXEMPLARY, OR PUNITIVE DAMAGES ARISING OUT OF 
* THE USE OR APPLICATION OF THE VERILOG model.  Further, 
* IDT reserves the right to make changes without notice to any
* product herein to improve reliability, function, or design.
* IDT does not convey any license under patent rights or
* any other intellectual property rights, including those of
* third parties. IDT is not obligated to provide maintenance 
* or support for the licensed Verilog model. 
*
****************************************************************/

//`define timing	// delay and timing check switch

//General constant
`timescale 1ns/10ps
`define t_format -9,1,"",1 // time format
`define a_width 17	// address width
`define e_width 2	// be number
`define b_width 9	// byte width
`define d_width 18	// data width
`define m_depth 131072	// memory depth
`define twait 0.1	// internal delay
`define tinbuf 1	// internal delay

//Read Cycle Parameters
`define Trc   10 // read cycle time
`define Taa   10 // address access time
`define Tace  10 // chip enable access time
`define Tabe   6 // byte enable access time
`define Taoe   5 // output enable access time
`define Toh    3 // output hold from address change time
`define Tlz    0 // output low Z time
`define Thz    4 // output high Z time

//Write Cycle Parameters
`define Twc   10 // write cycle time
`define Tew    8 // chip enable to end of write time
`define Taw    8 // address valid to end of write time
`define Tas    0 // address set up time
`define Twp    8 // write pulse width
`define Twr    0 // write recovery time
`define Tdw    6 // data valid to end of write time
`define Tdh    0 // data hold time
`define Tow    0 // output active from end of write time

//Busy Timing
`define Tbaa  10 // busy access time from address match
`define Tbda   7 // busy disable time from address not matched
`define Tbac   6 // busy access time from chip enable low
`define Tbdc   5 // busy access time from chip enable high
`define Taps   5 // arbitration priority setup time
`define Tbdd   1 // busy disable to valid data
`define Twb    0 // busy input to write
`define Twh    8 // write hold after busy
`define Twdd  22 // write pulse to data delay
`define Tddd  20 // write data valid to read data delay

//Interrupt Timing
`define Tins  10 // interrupt set time
`define Tinr  10 // interrupt recovery time

//Semaphore Timing
`define Tsop   4 // sem flag update pulse
`define Tswrd  5 // sem flag write to read time
`define Tsps   5 // sem flag contention window
`define Tsaa  10 // sem address access time

module idt70v639_top (laddr, ldata, lrw_, loe_, lbe_, lce0_, lce1, lsem_, lint_, lbusy_, loption, 
		raddr, rdata, rrw_, roe_, rbe_, rce0_, rce1, rsem_, rint_, rbusy_, roption, 
		ms_, tms, tck, tdi, tdo, trst);
input	[17-1:0]	laddr, raddr;
inout	[18-1:0]	ldata, rdata;
input	[2-1:0]	lbe_, rbe_;
input		lrw_, loe_, lce0_, lce1, lsem_, loption;
input		rrw_, roe_, rce0_, rce1, rsem_, roption;
output		lint_, rint_, tdo;
inout		lbusy_, rbusy_;
input		ms_, tms, tck, tdi, trst;

wire	[18-1:0]	ldata_chip, rdata_chip;
wire		lint_chip, rint_chip;
wire		lbusy_chip, rbusy_chip;

idt70v639_core core0	(laddr, ldata_chip, lrw_, loe_, lbe_, lce0_, lce1, lsem_, lint_chip, lbusy_chip, 
		 raddr, rdata_chip, rrw_, roe_, rbe_, rce0_, rce1, rsem_, rint_chip, rbusy_chip, ms_);

idt70v639_jtag jtag0	(laddr, ldata_chip, lrw_, loe_, lbe_, lce0_, lce1, lsem_, lint_chip, lbusy_chip, ldata, lint_, lbusy_, loption,
		 raddr, rdata_chip, rrw_, roe_, rbe_, rce0_, rce1, rsem_, rint_chip, rbusy_chip, rdata, rint_, rbusy_, roption,
		 ms_, tms, tck, tdi, tdo, trst);
endmodule

/***********************************************************************************************
#
# MODULE:	core
# PURPOSE:	memory chip core
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_core (laddr, ldata, lrw_, loe_, lbe_, lce0_, lce1, lsem_, lint_, lbusy_, 
		 raddr, rdata, rrw_, roe_, rbe_, rce0_, rce1, rsem_, rint_, rbusy_, ms_);
input	[17-1:0]	laddr, raddr;
inout	[18-1:0]	ldata, rdata;
input	[2-1:0]	lbe_, rbe_;
input		lrw_, loe_, lce0_, lce1, lsem_;
input		rrw_, roe_, rce0_, rce1, rsem_;
output		lint_, rint_;
inout		lbusy_, rbusy_;
input		ms_;

// Internal signal
reg	[17-1:0]	laddr_in, raddr_in;
reg	[18-1:0]	ldata_in, rdata_in;
wire	[2-1:0]	lwen, lerrflag, rwen, rerrflag;
wire		lce_, rce_;

initial
begin
	$timeformat(`t_format);
	laddr_in = 17'bz;
	raddr_in = 17'bz;
	ldata_in = 17'bz;
	rdata_in = 17'bz;
end

assign lce_ = !(!lce0_ & lce1);
assign rce_ = !(!rce0_ & rce1);

// Busy
idt70v639_busy busy0(laddr, lce_, lbusy_, lblkw_, raddr, rce_, rbusy_, rblkw_, ms_);

// Semaphore
idt70v639_sema semaphore0(laddr, ldata, lce_, lsem_, lrw_, loe_, raddr, rdata, rce_, rsem_, rrw_, roe_);

// Interrupt
idt70v639_int interrupt0(laddr, lce_, lrw_, loe_, lblkw_, lint_, raddr, rce_, rrw_, roe_, rblkw_, rint_);

// Memory
always @(laddr or lrw_) laddr_in <= #`tinbuf (!lrw_) ? laddr : 17'bz;
always @(ldata or lrw_) ldata_in <= #`tinbuf (!lrw_) ? ldata : 18'bz;
always @(raddr or rrw_) raddr_in <= #`tinbuf (!rrw_) ? raddr : 17'bz;
always @(rdata or rrw_) rdata_in <= #`tinbuf (!rrw_) ? rdata : 18'bz;

//byte0
reg [9-1:0] mem0[131072-1:0];
idt70v639_onebyte lb0(laddr, ldata[9-1:0], lrw_, loe_, lbe_[0], lce_, lblkw_, lwen[0], lerrflag[0], mem0[laddr], rdata[9-1:0], rrw_);
always @(negedge lwen[0]) if (lblkw_) #`twait mem0[laddr_in] <= lerrflag[0] ? 'hx : ldata_in[9-1:0];
idt70v639_onebyte rb0(raddr, rdata[9-1:0], rrw_, roe_, rbe_[0], rce_, rblkw_, rwen[0], rerrflag[0], mem0[raddr], ldata[9-1:0], lrw_);
always @(negedge rwen[0]) if (rblkw_) #`twait mem0[raddr_in] <= rerrflag[0] ? 'hx : rdata_in[9-1:0];
//byte0

//byte1
reg [9-1:0] mem1[131072-1:0];
idt70v639_onebyte lb1(laddr, ldata[9*2-1:9], lrw_, loe_, lbe_[1], lce_, lblkw_, lwen[1], lerrflag[1], mem1[laddr], rdata[9*2-1:9], rrw_);
always @(negedge lwen[1]) if (lblkw_) #`twait mem1[laddr_in] <= lerrflag[1] ? 'hx : ldata_in[9*2-1:9];
idt70v639_onebyte rb1(raddr, rdata[9*2-1:9], rrw_, roe_, rbe_[1], rce_, rblkw_, rwen[1], rerrflag[1], mem1[raddr], ldata[9*2-1:9], lrw_);
always @(negedge rwen[1]) if (rblkw_) #`twait mem1[raddr_in] <= rerrflag[1] ? 'hx : rdata_in[9*2-1:9];
//byte1

/*
reg [9-1:0] mem2[131072-1:0];
idt70v639_onebyte lb2(laddr, ldata[9*3-1:9*2], lrw_, loe_, lbe_[2], lce_, lblkw_, lwen[2], lerrflag[2], mem2[laddr], rdata[9*3-1:9*2], rrw_);
always @(negedge lwen[2]) if (lblkw_) #`twait mem2[laddr_in] <= lerrflag[2] ? 'hx : ldata_in[9*3-1:9*2];
idt70v639_onebyte rb2(raddr, rdata[9*3-1:9*2], rrw_, roe_, rbe_[2], rce_, rblkw_, rwen[2], rerrflag[2], mem2[raddr], ldata[9*3-1:9*2], lrw_);
always @(negedge rwen[2]) if (rblkw_) #`twait mem2[raddr_in] <= rerrflag[2] ? 'hx : rdata_in[9*3-1:9*2];
*/

/*
reg [9-1:0] mem3[131072-1:0];
idt70v639_onebyte lb3(laddr, ldata[9*4-1:9*3], lrw_, loe_, lbe_[3], lce_, lblkw_, lwen[3], lerrflag[3], mem3[laddr], rdata[9*4-1:9*3], rrw_);
always @(negedge lwen[3]) if (lblkw_) #`twait mem3[laddr_in] <= lerrflag[3] ? 'hx : ldata_in[9*4-1:9*3];
idt70v639_onebyte rb3(raddr, rdata[9*4-1:9*3], rrw_, roe_, rbe_[3], rce_, rblkw_, rwen[3], rerrflag[3], mem3[raddr], ldata[9*4-1:9*3], lrw_);
always @(negedge rwen[3]) if (rblkw_) #`twait mem3[raddr_in] <= rerrflag[3] ? 'hx : rdata_in[9*4-1:9*3];
*/
endmodule

/***********************************************************************************************
#
# MODULE:	onebyte
# PURPOSE:	one byte of the memory
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module	idt70v639_onebyte(addr, data, rw_, oe_, be_, ce_, blkw_, wen, errflag, mem, data_op, rw_op);
input	[17-1:0]	addr;
inout	[9-1:0]	data;
input	[9-1:0]	data_op;
input	[9-1:0]	mem;
input		rw_, oe_, be_, ce_, blkw_, rw_op;
output		wen, errflag;

// Local variable
wire		ren, wen, loop_back;
reg		rend;
reg	[2:0]	access;
reg	[9-1:0]	data_out, temp_op;
reg		err1, err2, err3, errflag;

`ifdef timing
specify
$setuphold(negedge wen, addr, `Taw, `Twr, err1);
$setuphold(negedge wen, data, `Tdw, `Tdh, err2);
$width(posedge wen, `Twp, 0, err3);
endspecify
always @(err1 or err2 or err3) errflag <= 1;
always @(posedge wen) errflag <= 0;
always @(negedge wen) if (!blkw_) $display($time, " Write operation is blocked due to busy");
parameter ptaa  = `Taa;
parameter ptace = `Tace;
parameter ptabe = `Tabe;
parameter ptaoe = `Taoe;
parameter pthz  = `Thz;
`else
parameter ptaa  = 0;
parameter ptace = 0;
parameter ptabe = 0;
parameter ptaoe = 0;
parameter pthz  = 0;
`endif

initial
begin
	rend = 0;
	access = 0;
	data_out = 9'bz;
	temp_op = 9'bz;
	err1 = 0;
	err2 = 0;
	err3 = 0;
	errflag = 0;
end

assign	ren = rw_ & !oe_ & !ce_ & !be_;
assign	wen = !rw_ & !ce_ & !be_;
assign	loop_back = rw_ & !rw_op & !blkw_;
always @(negedge ce_) access <= 1;
always @(negedge be_) access <= 2;
always @(negedge oe_) access <= 3;
always @(posedge rw_) access <= 4;
always @(addr) access <= 5;
always @(access) #`twait access <= 0;

always @(access)
if (rw_ & !oe_ & !ce_ & !be_)
case (access)
	1: data_out <= #(ptace) mem;
	2: data_out <= #(ptabe) mem;
	3: data_out <= #(ptaoe) mem;
	4: data_out <= #(ptace) mem;
	5: data_out <= #(ptaa) mem;
endcase
always @(negedge ren) data_out <= #pthz 9'bz;
always @(ren) #pthz rend <= ren;
always @(data_op) temp_op <= #pthz data_op;
assign data = rend ? (loop_back ? temp_op : data_out) : 9'bz;
endmodule

/***********************************************************************************************
#
# MODULE:	busy
# PURPOSE:	busy logic
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_busy(laddr, lce_, lbusy_, lblkw_,
            raddr, rce_, rbusy_, rblkw_, ms_);
input	[17-1:0]	laddr,raddr;
input		lce_, rce_, ms_;
inout		lbusy_, rbusy_;
output		lblkw_, rblkw_;

time		chg_laps, chg_raps;
reg		match, access;
reg		lbusy_out, rbusy_out;
reg		lblkw_out, rblkw_out;

`ifdef timing
parameter ptaps = `Taps;
parameter ptbaa = `Tbaa;
parameter ptbac = `Tbac;
parameter ptbda = `Tbda;
parameter ptbdc = `Tbdc;
`else
parameter ptaps = 0;
parameter ptbaa = 0;
parameter ptbac = 0;
parameter ptbda = 0;
parameter ptbdc = 0;
`endif

initial
begin
	chg_laps = 0;
	chg_raps = 0;
	match = 0;
	access = 0;
	lbusy_out = 1;
	rbusy_out = 1;
	lblkw_out = 1;
	rblkw_out = 1;
end

always @(laddr or lce_) chg_laps = $time;
always @(raddr or rce_) chg_raps = $time;
always @(laddr or raddr) access <= 0;
always @(lce_ or rce_) access <= 1;
always @(laddr or raddr or lce_ or rce_) match <= (!lce_ & !rce_) ? (laddr == raddr) : 0;
always @(posedge match)
	if (chg_laps <= chg_raps)
	begin
		if (chg_laps > chg_raps - ptaps) $display($time, " Arbitration priority setup time is not enough `Taps");
		rblkw_out <= 0;
		case(access)
			0: rbusy_out <= #ptbaa 0;
			1: rbusy_out <= #ptbac 0;
		endcase
	end
	else
	begin
		if (chg_laps < chg_raps + ptaps) $display($time, " Arbitration priority setup time is not enough `Taps");
		lblkw_out <= 0;
		case(access)
			0: lbusy_out <= #ptbaa 0;
			1: lbusy_out <= #ptbac 0;
		endcase
	end
always @(negedge match)
begin
	lblkw_out <= 1;
	rblkw_out <= 1;
	case(access)
		0: {lbusy_out,rbusy_out} <= #ptbda 2'b11;
		1: {lbusy_out,rbusy_out} <= #ptbdc 2'b11;
	endcase
end

assign lblkw_ = ms_ ? lblkw_out : lbusy_;
assign rblkw_ = ms_ ? rblkw_out : rbusy_;
assign lbusy_ = ms_ ? lbusy_out : 1'bz;
assign rbusy_ = ms_ ? rbusy_out : 1'bz;
endmodule

/***********************************************************************************************
#
# MODULE:	sema
# PURPOSE:	semaphore logic
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_sema(laddr, ldata, lce_, lsem_, lrw_, loe_,
                 raddr, rdata, rce_, rsem_, rrw_, roe_);
input [17-1:0] laddr, raddr;
inout [18-1:0] ldata, rdata;
input lce_, lsem_, lrw_, loe_;
input rce_, rsem_, rrw_, roe_;

reg	[17-1:0]	laddr_in, raddr_in;
reg	[18-1:0]	ldata_in, rdata_in;
reg	[18-1:0]	ldata_out, rdata_out;
reg	[8-1:0]	lsr, rsr, lrqst, rrqst;
reg	[2:0]	laccess, raccess;
reg		lrd, rrd;
wire		lw, lr, rw, rr;

initial
begin
	#`twait
	laccess = 0;
	raccess = 0;
	lsr = ~(8'b0);
	rsr = ~(8'b0);
	lrqst = ~(8'b0);
	rrqst = ~(8'b0);
	ldata_out = 18'bz;
	rdata_out = 18'bz;
	lrd = 0;
	rrd = 0;
end

`ifdef timing
specify
$setup(laddr, posedge lw, `Tas);
$setuphold(negedge lw, laddr, `Taw, `Twr);
$setuphold(negedge lw, ldata, `Tdw, `Tdh);
$width(posedge lw, `Twp);
$setup(raddr, posedge rw, `Tas);
$setuphold(negedge rw, raddr, `Taw, `Twr);
$setuphold(negedge rw, rdata, `Tdw, `Tdh);
$width(posedge rw, `Twp);
endspecify
parameter ptsaa = `Tsaa;
parameter ptace = `Tace;
parameter ptaoe = `Taoe;
parameter ptoh = `Toh;
`else
parameter ptsaa = 0;
parameter ptace = 0;
parameter ptaoe = 0;
parameter ptoh = 0;
`endif

always @(laddr or lrw_) laddr_in <= #`tinbuf (!lrw_) ? laddr : 17'bz;
always @(ldata or lrw_) ldata_in <= #`tinbuf (!lrw_) ? ldata : 18'bz;
assign lw = lce_ & !lsem_ & !lrw_;
assign lr = lce_ & !lsem_ & lrw_ & !loe_;
always @(posedge lce_) laccess = 1;
always @(negedge lsem_) laccess = 2;
always @(posedge lrw_) laccess = 3;
always @(negedge loe_) laccess = 4;
always @(laddr) laccess = 5;
always @(laccess) #`twait laccess = 0;
always @(negedge lw)
begin
	lrqst[laddr_in[2:0]] <= ldata_in[0];
	lsr[laddr_in[2:0]] <= ldata_in[0] ? 1 : !rsr[laddr_in[2:0]];
	rsr[laddr_in[2:0]] <= ldata_in[0] ? rrqst[laddr_in[2:0]] : rsr[laddr_in[2:0]];
end
always @(laccess)
if (lce_ & !lsem_ & lrw_ & !loe_)
case (laccess)
	1: ldata_out <= #ptace {18{lsr[laddr[2:0]]}};
	2: ldata_out <= #ptace {18{lsr[laddr[2:0]]}};
	3: ldata_out <= #ptace {18{lsr[laddr[2:0]]}};
	4: ldata_out <= #ptaoe {18{lsr[laddr[2:0]]}};
	5: ldata_out <= #ptsaa {18{lsr[laddr[2:0]]}};
endcase
always @(negedge lr) ldata_out <= #ptoh 18'bz;
always @(lr) #ptoh lrd <= lr;
assign ldata = lrd ? ldata_out : 18'bz;

always @(raddr or rrw_) raddr_in <= #`tinbuf (!rrw_) ? raddr : 17'bz;
always @(rdata or rrw_) rdata_in <= #`tinbuf (!rrw_) ? rdata : 18'bz;
assign rw = rce_ & !rsem_ & !rrw_;
assign rr = rce_ & !rsem_ & rrw_ & !roe_;
always @(posedge rce_) raccess = 1;
always @(negedge rsem_) raccess = 2;
always @(posedge rrw_) raccess = 3;
always @(negedge roe_) raccess = 4;
always @(raddr) raccess = 5;
always @(raccess) #`twait raccess = 0;
always @(negedge rw)
begin
	rrqst[raddr_in[2:0]] <= rdata_in[0];
	rsr[raddr_in[2:0]] <= rdata_in[0] ? 1 : !lsr[raddr_in[2:0]];
	lsr[raddr_in[2:0]] <= rdata_in[0] ? lrqst[raddr_in[2:0]] : lsr[raddr_in[2:0]];
end
always @(raccess)
if (rce_ & !rsem_ & rrw_ & !roe_)
case (raccess)
	1: rdata_out <= #ptace {18{rsr[raddr[2:0]]}};
	2: rdata_out <= #ptace {18{rsr[raddr[2:0]]}};
	3: rdata_out <= #ptace {18{rsr[raddr[2:0]]}};
	4: rdata_out <= #ptaoe {18{rsr[raddr[2:0]]}};
	5: rdata_out <= #ptsaa {18{rsr[raddr[2:0]]}};
endcase
always @(negedge rr) rdata_out <= #ptoh 18'bz;
always @(rr) #ptoh rrd <= rr;
assign rdata = rrd ? rdata_out : 18'bz;
endmodule

/***********************************************************************************************
#
# MODULE:	int
# PURPOSE:	interrupt logic
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_int(laddr, lce_, lrw_, loe_, lblkw_, lint_,
                 raddr, rce_, rrw_, roe_, rblkw_, rint_);
input [17-1:0] laddr, raddr;
input lce_, lrw_, loe_, lblkw_;
input rce_, rrw_, roe_, rblkw_;
output lint_, rint_;

reg		lint_, rint_;
wire		lwint, rwint, lrint, rrint;

initial
begin
	lint_ = 1;
	rint_ = 1;
end

`ifdef timing
specify
$setup(laddr, posedge lwint, `Tas);
$setup(laddr, posedge lrint, `Tas);
$hold(negedge lwint, laddr, `Twr);
$setup(raddr, posedge rwint, `Tas);
$setup(raddr, posedge rrint, `Tas);
$hold(negedge rwint, raddr, `Twr);
endspecify
parameter ptins  = `Tins;
parameter ptinr  = `Tinr;
`else
parameter ptins  = 0;
parameter ptinr  = 0;
`endif

assign lwint = !lce_ & !lrw_;
assign lrint = !lce_ & !loe_;
assign rwint = !rce_ & !rrw_;
assign rrint = !rce_ & !roe_;

always @(posedge lwint) if (lblkw_ & (laddr == {17{1'b1}})) rint_ <= #ptins 0;
always @(posedge rwint) if (rblkw_ & (raddr == {{17-1{1'b1}},1'b0})) lint_ <= #ptins 0;
always @(posedge lrint) if (lblkw_ & (laddr == {{17-1{1'b1}},1'b0})) lint_ <= #ptinr 1;
always @(posedge rrint) if (rblkw_ & (raddr == {17{1'b1}})) rint_ <= #ptinr 1;
endmodule

/***********************************************************************************************
#
# MODULE: 	jtag
# PURPOSE: 	jtag boundary scan chain
# DESIGNER:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_jtag (laddr, ldata_chip, lrw_, loe_, lbe_, lce0_, lce1, lsem_, lint_chip, lbusy_chip, ldata_jtag, lint_jtag, lbusy_jtag, loption,
		raddr, rdata_chip, rrw_, roe_, rbe_, rce0_, rce1, rsem_, rint_chip, rbusy_chip, rdata_jtag, rint_jtag, rbusy_jtag, roption,
		ms_, tms, tck, tdi, tdo, trst);
input	[17-1:0]	laddr, raddr;
inout	[18-1:0]	ldata_chip, rdata_chip, ldata_jtag, rdata_jtag;
input	[2-1:0]	lbe_, rbe_;
input		lrw_, loe_, lce0_, lce1, lsem_, loption;
input		rrw_, roe_, rce0_, rce1, rsem_, roption;
inout		lbusy_chip, rbusy_chip, lbusy_jtag, rbusy_jtag;
input		ms_, tms, tck, tdi, trst;
input		lint_chip, rint_chip;
output		lint_jtag, rint_jtag;
output		tdo;

wire		shiftdr, updatedr, clockdr, highz, jtoe, tdo_tap;
wire	[300:0]	so, di, do;
assign tdo = highz ? 1'bz : (jtoe ? tdo_tap : 1'bz);
idt70v639_tap tap0 (tck, trst, tms, tdi, tdo_tap, jtoe, so[0], tdi, shiftdr, updatedr, clockdr, extest, highz);   
idt70v639_oecell bscell_133(tck, tdi, lrw_, shiftdr, clockdr, updatedr, extest, so[133], do[133]);
idt70v639_oecell bscell_132(tck, so[133], lrw_, shiftdr, clockdr, updatedr, extest, so[132], do[132]);
idt70v639_incell bscell_131(tck, so[132], laddr[16], shiftdr, clockdr, so[131]);
idt70v639_incell bscell_130(tck, so[131], laddr[15], shiftdr, clockdr, so[130]);
idt70v639_incell bscell_129(tck, so[130], laddr[14], shiftdr, clockdr, so[129]);
idt70v639_incell bscell_128(tck, so[129], laddr[13], shiftdr, clockdr, so[128]);
idt70v639_incell bscell_127(tck, so[128], laddr[12], shiftdr, clockdr, so[127]);
idt70v639_incell bscell_126(tck, so[127], laddr[11], shiftdr, clockdr, so[126]);
idt70v639_incell bscell_125(tck, so[126], laddr[10], shiftdr, clockdr, so[125]);
idt70v639_incell bscell_124(tck, so[125], laddr[9], shiftdr, clockdr, so[124]);
idt70v639_incell bscell_123(tck, so[124], laddr[8], shiftdr, clockdr, so[123]);
idt70v639_incell bscell_122(tck, so[123], laddr[7], shiftdr, clockdr, so[122]);
idt70v639_incell bscell_121(tck, so[122], laddr[6], shiftdr, clockdr, so[121]);
idt70v639_incell bscell_120(tck, so[121], laddr[5], shiftdr, clockdr, so[120]);
idt70v639_incell bscell_119(tck, so[120], laddr[4], shiftdr, clockdr, so[119]);
idt70v639_incell bscell_118(tck, so[119], laddr[3], shiftdr, clockdr, so[118]);
idt70v639_incell bscell_117(tck, so[118], laddr[2], shiftdr, clockdr, so[117]);
idt70v639_incell bscell_116(tck, so[117], laddr[1], shiftdr, clockdr, so[116]);
idt70v639_incell bscell_115(tck, so[116], laddr[0], shiftdr, clockdr, so[115]);
idt70v639_incell bscell_114(tck, so[115], lce0_, shiftdr, clockdr, so[114]);
idt70v639_incell bscell_113(tck, so[114], lce1, shiftdr, clockdr, so[113]);
idt70v639_incell bscell_112(tck, so[113], lrw_, shiftdr, clockdr, so[112]);
idt70v639_incell bscell_111(tck, so[112], lsem_, shiftdr, clockdr, so[111]);
idt70v639_incell bscell_110(tck, so[111], loe_, shiftdr, clockdr, so[110]);
idt70v639_incell bscell_109(tck, so[110], lbe_[0], shiftdr, clockdr, so[109]);
idt70v639_incell bscell_108(tck, so[109], lbe_[1], shiftdr, clockdr, so[108]);
idt70v639_incell bscell_107(tck, so[108], loption, shiftdr, clockdr, so[107]);
idt70v639_oecell bscell_106(tck, so[107], !ms_, shiftdr, clockdr, updatedr, extest, so[106], do[106]);
idt70v639_incell bscell_105(tck, so[106], lbusy_jtag, shiftdr, clockdr, so[105]);
idt70v639_iocell bscell_104(tck, so[105], lbusy_chip, !do[106], shiftdr, clockdr, updatedr, extest, highz, so[104], lbusy_jtag);
idt70v639_outcell bscell_103(tck, so[104], lint_chip, shiftdr, clockdr, updatedr, extest, highz, so[103], lint_jtag);
idt70v639_oecell bscell_102(tck, so[103], rrw_, shiftdr, clockdr, updatedr, extest, so[102], do[102]);
idt70v639_oecell bscell_101(tck, so[102], rrw_, shiftdr, clockdr, updatedr, extest, so[101], do[101]);
idt70v639_incell bscell_100(tck, so[101], raddr[16], shiftdr, clockdr, so[100]);
idt70v639_incell bscell_99(tck, so[100], raddr[15], shiftdr, clockdr, so[99]);
idt70v639_incell bscell_98(tck, so[99], raddr[14], shiftdr, clockdr, so[98]);
idt70v639_incell bscell_97(tck, so[98], raddr[13], shiftdr, clockdr, so[97]);
idt70v639_incell bscell_96(tck, so[97], raddr[12], shiftdr, clockdr, so[96]);
idt70v639_incell bscell_95(tck, so[96], raddr[11], shiftdr, clockdr, so[95]);
idt70v639_incell bscell_94(tck, so[95], raddr[10], shiftdr, clockdr, so[94]);
idt70v639_incell bscell_93(tck, so[94], raddr[9], shiftdr, clockdr, so[93]);
idt70v639_incell bscell_92(tck, so[93], raddr[8], shiftdr, clockdr, so[92]);
idt70v639_incell bscell_91(tck, so[92], raddr[7], shiftdr, clockdr, so[91]);
idt70v639_incell bscell_90(tck, so[91], raddr[6], shiftdr, clockdr, so[90]);
idt70v639_incell bscell_89(tck, so[90], raddr[5], shiftdr, clockdr, so[89]);
idt70v639_incell bscell_88(tck, so[89], raddr[4], shiftdr, clockdr, so[88]);
idt70v639_incell bscell_87(tck, so[88], raddr[3], shiftdr, clockdr, so[87]);
idt70v639_incell bscell_86(tck, so[87], raddr[2], shiftdr, clockdr, so[86]);
idt70v639_incell bscell_85(tck, so[86], raddr[1], shiftdr, clockdr, so[85]);
idt70v639_incell bscell_84(tck, so[85], raddr[0], shiftdr, clockdr, so[84]);
idt70v639_incell bscell_83(tck, so[84], rce0_, shiftdr, clockdr, so[83]);
idt70v639_incell bscell_82(tck, so[83], rce1, shiftdr, clockdr, so[82]);
idt70v639_incell bscell_81(tck, so[82], rrw_, shiftdr, clockdr, so[81]);
idt70v639_incell bscell_80(tck, so[81], rsem_, shiftdr, clockdr, so[80]);
idt70v639_incell bscell_79(tck, so[80], roe_, shiftdr, clockdr, so[79]);
idt70v639_incell bscell_78(tck, so[79], ms_, shiftdr, clockdr, so[78]);
idt70v639_incell bscell_77(tck, so[78], rbe_[0], shiftdr, clockdr, so[77]);
idt70v639_incell bscell_76(tck, so[77], rbe_[1], shiftdr, clockdr, so[76]);
idt70v639_incell bscell_75(tck, so[76], roption, shiftdr, clockdr, so[75]);
idt70v639_incell bscell_74(tck, so[75], rbusy_jtag, shiftdr, clockdr, so[74]);
idt70v639_iocell bscell_73(tck, so[74], rbusy_chip, !do[106], shiftdr, clockdr, updatedr, extest, highz, so[73], rbusy_jtag);
idt70v639_outcell bscell_72(tck, so[73], rint_chip, shiftdr, clockdr, updatedr, extest, highz, so[72], rint_jtag);
idt70v639_incell bscell_71(tck, so[72], ldata_jtag[0], shiftdr, clockdr, so[71]);
idt70v639_iocell bscell_70(tck, so[71], ldata_chip[0], do[133], shiftdr, clockdr, updatedr, extest, highz, so[70], ldata_jtag[0]);
idt70v639_incell bscell_69(tck, so[70], rdata_jtag[0], shiftdr, clockdr, so[69]);
idt70v639_iocell bscell_68(tck, so[69], rdata_chip[0], do[102], shiftdr, clockdr, updatedr, extest, highz, so[68], rdata_jtag[0]);
idt70v639_incell bscell_67(tck, so[68], ldata_jtag[1], shiftdr, clockdr, so[67]);
idt70v639_iocell bscell_66(tck, so[67], ldata_chip[1], do[133], shiftdr, clockdr, updatedr, extest, highz, so[66], ldata_jtag[1]);
idt70v639_incell bscell_65(tck, so[66], rdata_jtag[1], shiftdr, clockdr, so[65]);
idt70v639_iocell bscell_64(tck, so[65], rdata_chip[1], do[102], shiftdr, clockdr, updatedr, extest, highz, so[64], rdata_jtag[1]);
idt70v639_incell bscell_63(tck, so[64], ldata_jtag[2], shiftdr, clockdr, so[63]);
idt70v639_iocell bscell_62(tck, so[63], ldata_chip[2], do[133], shiftdr, clockdr, updatedr, extest, highz, so[62], ldata_jtag[2]);
idt70v639_incell bscell_61(tck, so[62], rdata_jtag[2], shiftdr, clockdr, so[61]);
idt70v639_iocell bscell_60(tck, so[61], rdata_chip[2], do[102], shiftdr, clockdr, updatedr, extest, highz, so[60], rdata_jtag[2]);
idt70v639_incell bscell_59(tck, so[60], ldata_jtag[3], shiftdr, clockdr, so[59]);
idt70v639_iocell bscell_58(tck, so[59], ldata_chip[3], do[133], shiftdr, clockdr, updatedr, extest, highz, so[58], ldata_jtag[3]);
idt70v639_incell bscell_57(tck, so[58], rdata_jtag[3], shiftdr, clockdr, so[57]);
idt70v639_iocell bscell_56(tck, so[57], rdata_chip[3], do[102], shiftdr, clockdr, updatedr, extest, highz, so[56], rdata_jtag[3]);
idt70v639_incell bscell_55(tck, so[56], ldata_jtag[4], shiftdr, clockdr, so[55]);
idt70v639_iocell bscell_54(tck, so[55], ldata_chip[4], do[133], shiftdr, clockdr, updatedr, extest, highz, so[54], ldata_jtag[4]);
idt70v639_incell bscell_53(tck, so[54], rdata_jtag[4], shiftdr, clockdr, so[53]);
idt70v639_iocell bscell_52(tck, so[53], rdata_chip[4], do[102], shiftdr, clockdr, updatedr, extest, highz, so[52], rdata_jtag[4]);
idt70v639_incell bscell_51(tck, so[52], ldata_jtag[8], shiftdr, clockdr, so[51]);
idt70v639_iocell bscell_50(tck, so[51], ldata_chip[8], do[133], shiftdr, clockdr, updatedr, extest, highz, so[50], ldata_jtag[8]);
idt70v639_incell bscell_49(tck, so[50], rdata_jtag[8], shiftdr, clockdr, so[49]);
idt70v639_iocell bscell_48(tck, so[49], rdata_chip[8], do[102], shiftdr, clockdr, updatedr, extest, highz, so[48], rdata_jtag[8]);
idt70v639_incell bscell_47(tck, so[48], ldata_jtag[7], shiftdr, clockdr, so[47]);
idt70v639_iocell bscell_46(tck, so[47], ldata_chip[7], do[133], shiftdr, clockdr, updatedr, extest, highz, so[46], ldata_jtag[7]);
idt70v639_incell bscell_45(tck, so[46], rdata_jtag[7], shiftdr, clockdr, so[45]);
idt70v639_iocell bscell_44(tck, so[45], rdata_chip[7], do[102], shiftdr, clockdr, updatedr, extest, highz, so[44], rdata_jtag[7]);
idt70v639_incell bscell_43(tck, so[44], ldata_jtag[6], shiftdr, clockdr, so[43]);
idt70v639_iocell bscell_42(tck, so[43], ldata_chip[6], do[133], shiftdr, clockdr, updatedr, extest, highz, so[42], ldata_jtag[6]);
idt70v639_incell bscell_41(tck, so[42], rdata_jtag[6], shiftdr, clockdr, so[41]);
idt70v639_iocell bscell_40(tck, so[41], rdata_chip[6], do[102], shiftdr, clockdr, updatedr, extest, highz, so[40], rdata_jtag[6]);
idt70v639_incell bscell_39(tck, so[40], ldata_jtag[5], shiftdr, clockdr, so[39]);
idt70v639_iocell bscell_38(tck, so[39], ldata_chip[5], do[133], shiftdr, clockdr, updatedr, extest, highz, so[38], ldata_jtag[5]);
idt70v639_incell bscell_37(tck, so[38], rdata_jtag[5], shiftdr, clockdr, so[37]);
idt70v639_iocell bscell_36(tck, so[37], rdata_chip[5], do[102], shiftdr, clockdr, updatedr, extest, highz, so[36], rdata_jtag[5]);
idt70v639_incell bscell_35(tck, so[36], ldata_jtag[17], shiftdr, clockdr, so[35]);
idt70v639_iocell bscell_34(tck, so[35], ldata_chip[17], do[133], shiftdr, clockdr, updatedr, extest, highz, so[34], ldata_jtag[17]);
idt70v639_incell bscell_33(tck, so[34], rdata_jtag[17], shiftdr, clockdr, so[33]);
idt70v639_iocell bscell_32(tck, so[33], rdata_chip[17], do[102], shiftdr, clockdr, updatedr, extest, highz, so[32], rdata_jtag[17]);
idt70v639_incell bscell_31(tck, so[32], ldata_jtag[16], shiftdr, clockdr, so[31]);
idt70v639_iocell bscell_30(tck, so[31], ldata_chip[16], do[133], shiftdr, clockdr, updatedr, extest, highz, so[30], ldata_jtag[16]);
idt70v639_incell bscell_29(tck, so[30], rdata_jtag[16], shiftdr, clockdr, so[29]);
idt70v639_iocell bscell_28(tck, so[29], rdata_chip[16], do[102], shiftdr, clockdr, updatedr, extest, highz, so[28], rdata_jtag[16]);
idt70v639_incell bscell_27(tck, so[28], ldata_jtag[15], shiftdr, clockdr, so[27]);
idt70v639_iocell bscell_26(tck, so[27], ldata_chip[15], do[133], shiftdr, clockdr, updatedr, extest, highz, so[26], ldata_jtag[15]);
idt70v639_incell bscell_25(tck, so[26], rdata_jtag[15], shiftdr, clockdr, so[25]);
idt70v639_iocell bscell_24(tck, so[25], rdata_chip[15], do[102], shiftdr, clockdr, updatedr, extest, highz, so[24], rdata_jtag[15]);
idt70v639_incell bscell_23(tck, so[24], ldata_jtag[14], shiftdr, clockdr, so[23]);
idt70v639_iocell bscell_22(tck, so[23], ldata_chip[14], do[133], shiftdr, clockdr, updatedr, extest, highz, so[22], ldata_jtag[14]);
idt70v639_incell bscell_21(tck, so[22], rdata_jtag[14], shiftdr, clockdr, so[21]);
idt70v639_iocell bscell_20(tck, so[21], rdata_chip[14], do[102], shiftdr, clockdr, updatedr, extest, highz, so[20], rdata_jtag[14]);
idt70v639_incell bscell_19(tck, so[20], ldata_jtag[13], shiftdr, clockdr, so[19]);
idt70v639_iocell bscell_18(tck, so[19], ldata_chip[13], do[133], shiftdr, clockdr, updatedr, extest, highz, so[18], ldata_jtag[13]);
idt70v639_incell bscell_17(tck, so[18], rdata_jtag[13], shiftdr, clockdr, so[17]);
idt70v639_iocell bscell_16(tck, so[17], rdata_chip[13], do[102], shiftdr, clockdr, updatedr, extest, highz, so[16], rdata_jtag[13]);
idt70v639_incell bscell_15(tck, so[16], ldata_jtag[9], shiftdr, clockdr, so[15]);
idt70v639_iocell bscell_14(tck, so[15], ldata_chip[9], do[133], shiftdr, clockdr, updatedr, extest, highz, so[14], ldata_jtag[9]);
idt70v639_incell bscell_13(tck, so[14], rdata_jtag[9], shiftdr, clockdr, so[13]);
idt70v639_iocell bscell_12(tck, so[13], rdata_chip[9], do[102], shiftdr, clockdr, updatedr, extest, highz, so[12], rdata_jtag[9]);
idt70v639_incell bscell_11(tck, so[12], ldata_jtag[10], shiftdr, clockdr, so[11]);
idt70v639_iocell bscell_10(tck, so[11], ldata_chip[10], do[133], shiftdr, clockdr, updatedr, extest, highz, so[10], ldata_jtag[10]);
idt70v639_incell bscell_9(tck, so[10], rdata_jtag[10], shiftdr, clockdr, so[9]);
idt70v639_iocell bscell_8(tck, so[9], rdata_chip[10], do[102], shiftdr, clockdr, updatedr, extest, highz, so[8], rdata_jtag[10]);
idt70v639_incell bscell_7(tck, so[8], ldata_jtag[11], shiftdr, clockdr, so[7]);
idt70v639_iocell bscell_6(tck, so[7], ldata_chip[11], do[133], shiftdr, clockdr, updatedr, extest, highz, so[6], ldata_jtag[11]);
idt70v639_incell bscell_5(tck, so[6], rdata_jtag[11], shiftdr, clockdr, so[5]);
idt70v639_iocell bscell_4(tck, so[5], rdata_chip[11], do[102], shiftdr, clockdr, updatedr, extest, highz, so[4], rdata_jtag[11]);
idt70v639_incell bscell_3(tck, so[4], ldata_jtag[12], shiftdr, clockdr, so[3]);
idt70v639_iocell bscell_2(tck, so[3], ldata_chip[12], do[133], shiftdr, clockdr, updatedr, extest, highz, so[2], ldata_jtag[12]);
idt70v639_incell bscell_1(tck, so[2], rdata_jtag[12], shiftdr, clockdr, so[1]);
idt70v639_iocell bscell_0(tck, so[1], rdata_chip[12], do[102], shiftdr, clockdr, updatedr, extest, highz, so[0], rdata_jtag[12]);
endmodule

/***********************************************************************************************
#
# MODULE:	tap
# PURPOSE:	jtag tap controller
# DESIGNER:	Tom Diede
# MODIFIED:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_tap (jtag_clk_in,
            jtag_rst_in,
            jtag_tms_in,
            jtag_tdi_in,
            jtag_tdo_out,
            jtag_tdo_oe,

            boundary_so,
            is_so,

            shiftdr,
            updatedr,
            clockdr,

            extest,
            highz);   


input          jtag_clk_in;         // jtag clock 
input          jtag_rst_in;         // jtag reset
input          jtag_tms_in;         // jtag mode select
input          jtag_tdi_in;         // jtag data in
output         jtag_tdo_out;        // jtag data out
output         jtag_tdo_oe;         // active high tdo output enable

input          boundary_so;         // shift out from boundary scan reg (pad ring)
input          is_so;               // shift out from internal scan reg 

output         shiftdr;             // shift boundary scan reg
output         updatedr;            // update boundary scan reg
output         clockdr;             // clock the boundary scan reg

output         extest;    // puts boundary scan cells into boundary scan mode
output         highz;        // make all outputs highz




reg          tdo_q;
reg          jt_tdo_premux;
reg          jtag_tdo_out;
reg          jtag_tdo_oe;

reg          extest;
reg          shiftdr;
reg          updatedr;
reg          shiftid;
reg          captureid;
reg          update_q;
reg          shift_q;
reg          capture_q;
reg          shiftby;
reg          captureby;
reg          shiftir;
reg          captureir;
reg          bypass_reg;

reg          capturedr;

reg          highz_q;
reg          highz_q_reg;
reg          highz_latch;

reg  [2:0]   sel_reg_q;          
reg  [2:0]   sel_reg_reg;
reg  [2:0]   sel_reg_latch;

reg          mode_q;
reg          mode_q_reg;

reg  [3:0]   jt_state;           // state register
reg  [3:0]   jt_state_latch;     // clocked on falling edge of jtag_clk

reg  [3:0]   ir_reg;             // instruction register
reg  [3:0]   ir_reg_latch;       // instruction register output latch,  clocked on falling edge of jtag_clk

reg  [31:0]  jt_id_reg;
reg          id_so;


wire  highz = highz_latch;
wire  tdo_oe_q = (shift_q) ? 1'b1 : 1'b0;
wire  clockdr  = shiftdr || capturedr;



parameter
    test_logic_reset = 4'b1111,
    run_test_idle    = 4'b1100,
    select_dr_scan   = 4'b0111,
    capture_dr       = 4'b0110,
    shift_dr         = 4'b0010,
    exit1_dr         = 4'b0001,
    pause_dr         = 4'b0011,
    exit2_dr         = 4'b0000,
    update_dr        = 4'b0101,
    select_ir_scan   = 4'b0100,
    capture_ir       = 4'b1110,
    shift_ir         = 4'b1010,
    exit1_ir         = 4'b1001,
    pause_ir         = 4'b1011,
    exit2_ir         = 4'b1000,
    update_ir        = 4'b1101;



// id register
   always @(jt_id_reg) id_so = jt_id_reg[0];


// Registers and state bits
always @(posedge jtag_clk_in or posedge jtag_rst_in)
  if (jtag_rst_in)
    begin
      highz_q_reg <= 1'b0;
      sel_reg_reg    <= 3'b001;                   // select device id register
      mode_q_reg     <= 1'b0;
      bypass_reg     <= 1'b0;       
      ir_reg         <= 4'b0010;                  // device id always selected on reset
      jt_id_reg      <= 32'h30c067;
      jt_state       <= 4'b1111;                  // test_logic_reset
    end
  else
    begin

      highz_q_reg <= highz_latch;
      sel_reg_reg    <= sel_reg_latch;
      mode_q_reg     <= extest;

      casez (sel_reg_latch)   
        3'b000:                             // bypass register
          begin                   
            ir_reg <= ir_reg_latch;
            if (captureby)    bypass_reg <= 1'b0;        // always load 0 during capture bypass
            else if (shiftby) bypass_reg <= jtag_tdi_in;
            else              bypass_reg <= bypass_reg;
          end
        3'b011:                             // instruction register
          begin
            bypass_reg <= bypass_reg;
            if (shiftir)
              begin
                ir_reg[2:0] <= ir_reg_latch[3:1];
                ir_reg[3]   <= jtag_tdi_in;
              end
            else if (captureir)
              begin
                ir_reg <= 4'b1101;          // as per specification
              end
            else
              begin
                ir_reg <= ir_reg_latch;
              end
           end
        default:
           begin
             bypass_reg <= bypass_reg;
             ir_reg     <= ir_reg_latch;
           end
      endcase

// tap state machine 
     casez (jt_state_latch)
       test_logic_reset: 
         jt_state <= jtag_tms_in ? test_logic_reset : run_test_idle;
       run_test_idle:
         jt_state <= jtag_tms_in ? select_dr_scan : run_test_idle;
       select_dr_scan:
         jt_state <= jtag_tms_in ? select_ir_scan : capture_dr;
       capture_dr:
         jt_state <= jtag_tms_in ? exit1_dr : shift_dr;
       shift_dr:
         jt_state <= jtag_tms_in ? exit1_dr : shift_dr;
       exit1_dr:
         jt_state <= jtag_tms_in ? update_dr : pause_dr;
       pause_dr:
         jt_state <= jtag_tms_in ? exit2_dr : pause_dr;
       exit2_dr:
         jt_state <= jtag_tms_in ? update_dr : shift_dr;
       update_dr:
         jt_state <= jtag_tms_in ? select_dr_scan : run_test_idle;
       select_ir_scan:
         jt_state <= jtag_tms_in ? test_logic_reset : capture_ir;
       capture_ir:
         jt_state <= jtag_tms_in ? exit1_ir : shift_ir;
       shift_ir:
         jt_state <= jtag_tms_in ? exit1_ir : shift_ir;
       exit1_ir:
         jt_state <= jtag_tms_in ? update_ir : pause_ir;
       pause_ir:
         jt_state <= jtag_tms_in ? exit2_ir : pause_ir;
       exit2_ir:
         jt_state <= jtag_tms_in ? update_ir : shift_ir;
       update_ir:
         jt_state <= jtag_tms_in ? select_dr_scan : run_test_idle;
       default:
         jt_state <= 4'bxxxx;
     endcase


     // jtag id code
     casez ({captureid, shiftid})
        2'b01: begin                               // shift id
                 jt_id_reg[30:0] <= jt_id_reg[31:1];
                 jt_id_reg[31]   <= jtag_tdi_in;
               end
        2'b10: begin                               // capture id
                     jt_id_reg[31:0] <= 32'h30c067;
               end
        default: jt_id_reg <= jt_id_reg;
     endcase
 end





// State decodes
always @(highz_q_reg or sel_reg_reg or bypass_reg or id_so or ir_reg or id_so or jt_state or mode_q_reg)

   begin
         casez (sel_reg_reg)   
           3'b000:                   // bypass register
               tdo_q = bypass_reg;

           3'b001:                   // ID register
               tdo_q = id_so;

           3'b010:                   // boundary scan register 
               tdo_q = 1'b1;

           3'b011:                   // instruction register
               tdo_q = ir_reg[0];

           3'b100:                   // internal scan register
               tdo_q = id_so;

           default:
               tdo_q = bypass_reg;
         endcase


         highz_q = highz_q_reg;
         mode_q     = mode_q_reg;
         sel_reg_q  = sel_reg_reg;

         casez (jt_state) 
             4'b1111:     // test_logic_reset
               begin
                 update_q   = 1'b0; 
                 shift_q    = 1'b0;
                 capture_q  = 1'b0;
                 mode_q     = 1'b0;
                 sel_reg_q  = 3'b001;       // select read-id register
                 highz_q = 1'b0;
               end

             4'b1100:     // run_test_idle  // instructions are executed here
               begin
                 update_q  = 1'b0; 
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end

             4'b?110:      // (jt_state == capture_dr || jt_state == capture_ir)
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b1;
               end

             4'b?010:      // (jt_state == shift_dr || jt_state == shift_ir)
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b1;
                 capture_q = 1'b0;
                 mode_q    = mode_q_reg;
               end
 
             4'b0101:      // (jt_state == update_dr)
               begin
                 update_q  = 1'b1;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end
      
             4'b1101:      // (jt_state == update_ir)
               begin
                 update_q   = 1'b1;
                 shift_q    = 1'b0;
                 capture_q  = 1'b0;
                 highz_q = 1'b0;

                 casez (ir_reg)          // decode instruction reg
                    4'b0000:                          // command = extest 
                      begin
                        sel_reg_q = 3'b010;           // select boundary scan register
                        mode_q    = 1'b1;             // enable boundary scan mode
                      end


                    4'b0001:                          // command = sample/preload
                      begin
                        sel_reg_q = 3'b010;           // select boundary scan register
                        mode_q    = 1'b0;             // de-select boundary scan mode
                      end


                    4'b0010:                          // command = idcode
                        sel_reg_q = 3'b001;           // select id register


                    4'b0100:                          // command = highz
                      begin
                        highz_q = 1'b1;            // highz all outputs
                        mode_q     = 1'b1;            // enable boundary scan mode
                        sel_reg_q  = 3'b000;          // bypass
                      end

                    4'b0011:                          // command = clamp
                      begin
                        sel_reg_q = 3'b000;           // select bypass register
                        mode_q    = 1'b1;             // enable boundary scan mode
                      end

                    4'b0110:                          // command = fuse option programming
                      begin
                        sel_reg_q  = 3'b110;          // select fuse option register
                        mode_q     = 1'b0;            // disable boundary scan mode
                      end

                    4'b1111:                          // command = bypass
                      begin
                        sel_reg_q = 3'b000;           // select bypass register
                        mode_q    = 1'b0;             // disable boundary scan mode
                      end

                    4'b1101: ;                        // command = validate IR,  don't need to do anything

                    4'b0101: ;                        // reserved
                    4'b0111: ;                        // reserved
                    4'b1000: ;                        // reserved
                    4'b1001: ;                        // reserved
                    4'b1010: ;                        // reserved
                    4'b1011: ;                        // reserved
                    4'b1100: ;                        // reserved
                    4'b1110: ;                        // reserved

                 endcase
               end
                   
             4'b0100:      // instruction register path select-IR-Scan
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
                 sel_reg_q = 3'b011;                  // select instruction register
               end
   
             4'b0111:     // boundary scan register path - select-DR-scan
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end
   
             4'b?001:     // Exit1-DR || Exit1-IR
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end
   
             4'b?011:     // Pause-DR || Pause-IR
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end
   
             4'b?000:    // Exit2-DR || Exit2-IR
               begin
                 update_q  = 1'b0;
                 shift_q   = 1'b0;
                 capture_q = 1'b0;
               end
         endcase   
   end
   



always @(sel_reg_q or update_q or jt_state)
   begin
      casez (sel_reg_q)
         3'b000: updatedr = 1'b0;                                        // bypass register
         3'b001: updatedr = 1'b0;                                        // read id register
         3'b010: updatedr = (jt_state == update_dr) ? update_q : 1'b0;   // boundary scan register
         3'b011: updatedr = 1'b0;                                        // instruction register
         3'b110: updatedr = 1'b0;                                        // internal scan register
         default : updatedr = 1'b0;
      endcase
   end



always @(sel_reg_latch or jt_tdo_premux or boundary_so or is_so)
   begin
      casez (sel_reg_latch)
        3'b000:                  // bypass register
            jtag_tdo_out = jt_tdo_premux;
        3'b001:                  // id register
            jtag_tdo_out = jt_tdo_premux;
        3'b010:                  // boundary scan register
            jtag_tdo_out = boundary_so;
        3'b011:                  // instruction register
            jtag_tdo_out = jt_tdo_premux;
        3'b110:                  // internal scan test
            jtag_tdo_out = is_so;
        default:
            jtag_tdo_out = jt_tdo_premux;
      endcase
   end



  always @(negedge jtag_clk_in or posedge jtag_rst_in)
    if (jtag_rst_in)
      begin
        jt_state_latch   <= 4'b1111;
        ir_reg_latch     <= 4'b0010;     // device id selected on reset
        highz_latch   <= 1'b0;
        sel_reg_latch    <= 3'b001;      // read-id
        extest <= 1'b0;
        jtag_tdo_oe      <= 1'b0;
        jt_tdo_premux    <= 1'b0;
      end
    else
      begin
        jt_state_latch   <= jt_state;
        ir_reg_latch     <= ir_reg;
        highz_latch   <= highz_q;
        sel_reg_latch    <= sel_reg_q;
        extest <= mode_q;
        jtag_tdo_oe      <= tdo_oe_q;
        jt_tdo_premux    <= tdo_q;
      end
   


  always @(negedge jtag_clk_in or posedge jtag_rst_in)
    if (jtag_rst_in)
      begin
        captureir <= 1'b0;
        shiftir   <= 1'b0;
        captureid <= 1'b0;
        shiftid   <= 1'b0;
        capturedr <= 1'b0;
        shiftdr   <= 1'b0;      // module output
        captureby <= 1'b0;
        shiftby   <= 1'b0;
      end
    else
      begin
        casez (sel_reg_q)               	 
          3'b000:			// bypass register
            begin
              captureby <= capture_q;
              shiftby   <= shift_q;
              captureid <= 1'b0;
              shiftid   <= 1'b0;
              capturedr <= 1'b0;
              shiftdr   <= 1'b0;
              captureir <= 1'b0;
              shiftir   <= 1'b0;
            end
          3'b001:			// id register
            begin
              captureby <= 1'b0;
              shiftby   <= 1'b0;
              captureid <= capture_q;
              shiftid   <= shift_q;
              capturedr <= 1'b0;
              shiftdr   <= 1'b0;
              captureir <= 1'b0;
              shiftir   <= 1'b0;
            end
          3'b010: 			// boundary scan register 
            begin
              captureby <= 1'b0;
              shiftby   <= 1'b0;
              captureid <= 1'b0;
              shiftid   <= 1'b0;
              capturedr <= capture_q;
              shiftdr   <= shift_q;
              captureir <= 1'b0;
              shiftir   <= 1'b0;
            end
          3'b011: 			// instruction register
            begin
              captureby <= 1'b0;
              shiftby   <= 1'b0;
              captureid <= 1'b0;
              capturedr <= 1'b0;
              shiftdr   <= 1'b0;
              captureir <= capture_q;
              shiftir   <= shift_q;
            end
          3'b110:                      // internal scan register
            begin
              captureby <= 1'b0;
              shiftby   <= 1'b0;
              captureid <= 1'b0;
              capturedr <= 1'b0;
              shiftdr   <= 1'b0;
              captureir <= 1'b0;
              shiftir   <= 1'b0;
            end
          default :
            begin
              captureby <= 1'bx;
              shiftby   <= 1'bx;
              captureid <= 1'bx;
              shiftid   <= 1'bx;
              capturedr <= 1'bx;
              shiftdr   <= 1'bx;
              captureir <= 1'bx;
              shiftir   <= 1'bx;
            end
        endcase
      end

endmodule

// si - boundary scan serial input     
// so - boundary scan serial output
// di - data from internal logic
// do - data to io block
// pi - data from io block

/***********************************************************************************************
#
# MODULE:	incell
# PURPOSE:	jtag boundary scan input cell
# DESIGNER:	Tom Diede
# MODIFIED:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_incell (jtag_clk_in, si, di, shiftdr, clockdr, so);
input   jtag_clk_in;
input   si;
input   di;
input   shiftdr;
input   clockdr;
output  so; 

reg so_internal; 
reg so; 
wire n1 = shiftdr ? si : di;

initial
begin
	so = 0;
	so_internal = 0;
end

always @(posedge jtag_clk_in) so_internal <= clockdr ? n1 : so_internal;
always @(negedge jtag_clk_in) so <= so_internal;
endmodule

/***********************************************************************************************
#
# MODULE:	iocell
# PURPOSE:	jtag boundary scan io cell
# DESIGNER:	Tom Diede
# MODIFIED:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_iocell (jtag_clk_in, si, dio, oe, shiftdr, clockdr, updatedr, extest, highz, so, pio);
input   jtag_clk_in;
inout   pio;
input   oe;
input   si;
inout   dio;
input   shiftdr;
input   clockdr;
input   updatedr;
input   extest;
input	highz;
output  so;

reg so_internal;
reg so;
reg n2;
wire n1 = shiftdr ? si : dio;
wire pio = highz ? 'hz : (extest ? n2 : (oe ? dio : 'hz));
wire dio = extest ? 'hz : (oe ? 'hz : pio);

initial
begin
	n2 = 0;
	so = 0;
	so_internal = 0;
end

always @(posedge jtag_clk_in) so_internal <= clockdr ? n1 : so_internal;
always @(negedge jtag_clk_in)
begin
    n2 <= updatedr ? so : n2;
    so <= so_internal;
end
endmodule

/***********************************************************************************************
#
# MODULE:	outcell
# PURPOSE:	jtag boundary scan output cell
# DESIGNER:	Tom Diede
# MODIFIED:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_outcell (jtag_clk_in, si, di, shiftdr, clockdr, updatedr, extest, highz, so, do);
input   jtag_clk_in;
input   si;
input   di;
input   shiftdr;
input   clockdr;
input   updatedr;
input   extest;
input	highz;
output  so;
output  do; 

reg so_internal;
reg so;
reg n2;
wire n1 = shiftdr ? si : di;
wire do = highz ? 1'bz : (extest ? n2 : di);

initial
begin
	n2 = 0;
	so = 0;
	so_internal = 0;
end

always @(posedge jtag_clk_in) so_internal <= clockdr ? n1 : so_internal;
always @(negedge jtag_clk_in)
begin
    n2 <= updatedr ? so : n2;
    so <= so_internal;
end
endmodule

/***********************************************************************************************
#
# MODULE:	oecell
# PURPOSE:	jtag boundary scan output enable control cell
# DESIGNER:	Tom Diede
# MODIFIED:	Chenhao Geng
#
***********************************************************************************************/

module idt70v639_oecell (jtag_clk_in, si, di, shiftdr, clockdr, updatedr, extest, so, do);
input   jtag_clk_in;
input   si;
input   di;
input   shiftdr;
input   clockdr;
input   updatedr;
input   extest;
output  so;
output  do; 

reg so_internal;
reg so;
reg n2;

initial
begin
	n2 = 0;
	so = 0;
	so_internal = 0;
end

wire n1 = shiftdr ? si : di;
wire do = extest ? n2 : di;

always @(posedge jtag_clk_in)
begin
    so_internal <= clockdr ? n1 : so_internal;
end

always @(negedge jtag_clk_in)
begin
    n2 <= updatedr ? so : n2;
    so <= so_internal;
end
endmodule
